from ... import Subject


class SubjectMNI(Subject):
    """Atlases from the Montreal Neurological Institute (MNI).

    See `the website <http://nist.mni.mcgill.ca/?page_id=714>`_
    for more information.
    """
    url_base = 'http://packages.bic.mni.mcgill.ca/mni-models/'
